unit gfx_png;
{***********************************************************************
Unit gfx_png.PAS v1.0  0400
    (c) by Andreas Moser, amoser@amoser.de,

    Delphi version : Delphi  4

    gfx_png is part of the gfx_library collection

    You may use this sourcecode for your freewareproducts.
    You may modify this source-code for your own use.
    You may recompile this source-code for your own use.

    All functions, procedures and classes may NOT be used in commercial
    products without the permission of the author. For parts of this library
    not written by me, you have to ask for permission by their
    respective authors.

    Disclaimer of warranty: "This software is supplied as is. The author
    disclaims all warranties, expressed or implied, including, without
    limitation, the warranties of merchantability and of fitness for any
    purpose. The author assumes no liability for damages, direct or
    consequential, which may result from the use of this software."

    All brand and product names are marks or registered marks of their
    respective companies.

    Please report bugs to:
    Andreas Moser  amoser@amoser.de

    The PNG converter algorythms are based on PNGImage by Edmund H. Hand (edhand@worldnet.att.net)
    and are using the LIBPNG.DLL

********************************************************************************}
interface
uses SysUtils,
     classes,
     Windows,
     JPEG,
     Graphics,
     io_files,
     gfx_errors,
     gfx_basedef,
     pnglib;

// -----------------------------------------------------------------------------
//
//			PNG Classes
//
// -----------------------------------------------------------------------------

type
  TPNGBitmap = class(TBitmap)
  public
    procedure LoadFromStream(Stream: TStream);override;
  end;

type
  TPNGFile = class(TObject)
    FBitmap:        TBitmap;
    FBitDepth:      Integer;
    FBytesPerPixel: Integer;
    FColorType:     Integer;
    FHeight:        Integer;
    FWidth:         Integer;

    Data:           PByte;
    RowPtrs:        PByte;
    RefCount:       Integer;

    procedure InitializeDemData;
    procedure TransferPNGtoBitmap;
  public
    constructor Create;
    destructor Destroy;override;
    procedure LoadFromFile(filename: String);
    procedure LoadFromStream(STream:TStream);
    procedure SaveToFile(const Filename: string);
    property Bitmap:TBitmap read FBitmap write FBitmap;
end;


implementation

//***********************************************************
//
// TPNGFile
//
//***********************************************************

// -----------------------------------------------------------------------------
//
//			Create
//
// -----------------------------------------------------------------------------
constructor TPNGFile.Create;
begin
  Data     := nil;
  RowPtrs  := nil;
  FHeight  := 0;
  FWidth   := 0;
  RefCount := 1;
  FBitmap:=TBitmap.Create;
end;

// -----------------------------------------------------------------------------
//
//			Destroy
//
// -----------------------------------------------------------------------------
destructor TPNGFile.Destroy;
begin
  if Data <> nil then
    FreeMem(Data);
  if RowPtrs <> nil then
    FreeMem(RowPtrs);
  FBitmap.Free;
  inherited;
end;

// -----------------------------------------------------------------------------
//
//			LoadFromFile
//
// -----------------------------------------------------------------------------
procedure TPNGFile.LoadFromFile(filename: String);
var
  png:      PPng_Struct;
  pnginfo:  PPng_Info;
  rowbytes: Cardinal;
  PngFile:  Pointer;
  tmp:      array[0..32] of char;
begin
  pngfile := png_open_file(@Filename[1], 'rb');
  if pngfile = nil then
  begin
    if GFXRaiseErrors then raise Exception.Create('Error opening PNG-file ' + Filename + '!');
    GFXFileErrorList.Add('Error opening PNG-file ' + Filename + '!')
  end;
  try
    StrPCopy(tmp, PNG_LIBPNG_VER_STRING);
    try
      png := png_create_read_struct(tmp, nil, nil, nil);
      if png <> nil then
      begin
        try
          pnginfo := png_create_info_struct(png);
          png_init_io(png, pngfile);
          png_set_read_status_fn(png, nil);
          png_read_info(png, pnginfo);
          png_get_IHDR(png, pnginfo, @FWidth, @FHeight,
                       @FBitDepth, @FColorType, nil, nil, nil);
          png_set_invert_alpha(png);
          // if bit depth is less than 8 then expand...
          if (FColorType = PNG_COLOR_TYPE_PALETTE) and
             (FBitDepth <= 8) then
            png_set_expand(png);
          if (FColorType = PNG_COLOR_TYPE_GRAY) and
             (FBitDepth < 8) then
            png_set_expand(png);
          // Add alpha channel if pressent
          if png_get_valid(png, pnginfo, PNG_INFO_tRNS) = UINT(PNG_INFO_tRNS) then //changed to UINT
            png_set_expand(png);
          // expand images to 1 pixel per byte
          if FBitDepth < 8 then
            png_set_packing(png);
          // Swap 16 bit images to PC Format
          if FBitDepth = 16 then
            png_set_swap(png);
          // update the info structure
          png_read_update_info(png, pnginfo);
          png_get_IHDR(png, pnginfo, @FWidth, @FHeight,
                       @FBitDepth, @FColorType, nil, nil, nil);

          rowbytes := png_get_rowbytes(png, pnginfo);
          FBytesPerPixel := rowbytes div cardinal (FWidth);  //changed to UINT
          InitializeDemData;
          if (Data <> nil) and (RowPtrs <> nil) then
            // Read the image
            png_read_image(png, PPByte(RowPtrs));
        finally
          png_destroy_read_struct(@png, @pnginfo, nil);
        end;  // try pnginfo create
      end;  // png <> nil
    except
      if GFXRaiseErrors then  Exception.Create('Error reading PNG-file!');
      GFXFileErrorList.Add('Error reading PNG-file!');
    end;  // try png create

  finally
    png_close_file(pngfile);
    TransferPNGtoBitmap;
  end;
end;


// -----------------------------------------------------------------------------
//
//			LoadFromStream
//
// -----------------------------------------------------------------------------
procedure TPNGFile.LoadFromStream(Stream:TStream);
var
  png:      PPng_Struct;
  pnginfo:  PPng_Info;
  rowbytes: Cardinal;
  PngFile:  Pointer;
  tmp:      array[0..32] of char;
  aStream:TMemoryStream;
begin
  aStream:=TMemoryStream.Create;
  aStream.LoadFromStream(Stream);
  try
  aStream.SaveToFile('~tmp~.png');
  pngfile := png_open_file('~tmp~.png', 'rb');
  if pngfile = nil then
  begin
    if GFXRaiseErrors then raise Exception.Create('Error opening PNG-file !');
    GFXFileErrorList.Add('Error opening PNG-file !')
  end;
  try
    StrPCopy(tmp, PNG_LIBPNG_VER_STRING);
    try
      png := png_create_read_struct(tmp, nil, nil, nil);
      if png <> nil then
      begin
        try
          pnginfo := png_create_info_struct(png);
          png_init_io(png, pngfile);
          png_set_read_status_fn(png, nil);
          png_read_info(png, pnginfo);
          png_get_IHDR(png, pnginfo, @FWidth, @FHeight,
                       @FBitDepth, @FColorType, nil, nil, nil);
          png_set_invert_alpha(png);
          // if bit depth is less than 8 then expand...
          if (FColorType = PNG_COLOR_TYPE_PALETTE) and
             (FBitDepth <= 8) then
            png_set_expand(png);
          if (FColorType = PNG_COLOR_TYPE_GRAY) and
             (FBitDepth < 8) then
            png_set_expand(png);
          // Add alpha channel if pressent
          if png_get_valid(png, pnginfo, PNG_INFO_tRNS) = UINT(PNG_INFO_tRNS) then //changed to UINT
            png_set_expand(png);
          // expand images to 1 pixel per byte
          if FBitDepth < 8 then
            png_set_packing(png);
          // Swap 16 bit images to PC Format
          if FBitDepth = 16 then
            png_set_swap(png);
          // update the info structure
          png_read_update_info(png, pnginfo);
          png_get_IHDR(png, pnginfo, @FWidth, @FHeight,
                       @FBitDepth, @FColorType, nil, nil, nil);

          rowbytes := png_get_rowbytes(png, pnginfo);
          FBytesPerPixel := rowbytes div cardinal (FWidth);  //changed to UINT
          InitializeDemData;
          if (Data <> nil) and (RowPtrs <> nil) then
            // Read the image
            png_read_image(png, PPByte(RowPtrs));
        finally
          png_destroy_read_struct(@png, @pnginfo, nil);
        end;  // try pnginfo create
      end;  // png <> nil
    except
      if GFXRaiseErrors then  Exception.Create('Error reading PNG-file!');
      GFXFileErrorList.Add('Error reading PNG-file!');
    end;  // try png create

  finally
    png_close_file(pngfile);
    TransferPNGtoBitmap;
  end;
  finally
  aStream.Free;
  DeleteFile(PChar('~tmp~.png'));
  end;
end;

// -----------------------------------------------------------------------------
//
//			LoadFromStream
//
// -----------------------------------------------------------------------------
procedure TPNGBitmap.LoadFromStream(Stream:TStream);
var
  aPNG: TPNGFile;
  aStream: TMemoryStream;
begin
  aPNG := TPNGFile.Create;
  try
    aPNG.LoadFromStream(Stream);
    aStream := TMemoryStream.Create;
    try
      aPNG.Bitmap.SaveToStream(aStream);
      aStream.Position:=0;
      inherited LoadFromStream(aStream);
    finally
      aStream.Free;
    end;
  finally
    aPNG.Free;
  end;
end;


// -----------------------------------------------------------------------------
//
//			TransferPNGToBitmap
//
// -----------------------------------------------------------------------------
procedure TPNGFile.TransferPNGtoBitmap;
var
  valuep:  PByte;
  x, y:    Integer;
  ndx:     Integer;
  sl:      PByteArray;  // Scanline of bitmap
  slbpp:   Integer;     // Scanline bytes per pixel
  a, r, g, b: Byte;
begin
  FBitmap.Height := FHeight;
  FBitmap.Width  := FWidth;
  case FBytesPerPixel of
    2: begin
      FBitmap.PixelFormat := pf16Bit;
      slbpp := 2;
    end;
    else begin
      FBitmap.PixelFormat := pf24Bit;
      slbpp := 3;
    end;
  end;
  // point to data
  valuep := Data;
  for y := 0 to FHeight - 1 do
  begin
    sl := FBitmap.Scanline[y];  // current scanline
    for x := 0 to FWidth - 1 do
    begin
      ndx := x * slbpp;    // index into current scanline
      if FBytesPerPixel = 2 then
      begin
        // handle 16bit grayscale images, this will display them
        // as a 16bit color image, kinda hokie but fits my needs
        // without altering the data.
        sl[ndx]     := valuep^;  Inc(valuep);
        sl[ndx + 1] := valuep^;  Inc(valuep);
      end
      else if FBytesPerPixel = 3 then
      begin
        // RGB - swap blue and red for windows format
        sl[ndx + 2] := valuep^;  Inc(valuep);
        sl[ndx + 1] := valuep^;  Inc(valuep);
        sl[ndx]     := valuep^;  Inc(valuep);
      end
      else  // 4 bytes per pixel of image data
      begin
        // Alpha chanel present and RGB
        // this is what PNG is all about
        r := valuep^;  Inc(valuep);
        g := valuep^;  Inc(valuep);
        b := valuep^;  Inc(valuep);
        a := valuep^;  Inc(valuep);
        if a = 0 then
        begin
          // alpha is zero so no blending, just image data
          sl[ndx]     := b;
          sl[ndx + 1] := g;
          sl[ndx + 2] := r;
        end
        else if a < 255 then
        begin
          // blend with data from ACanvas as background
          sl[ndx]     := ((sl[ndx] * a) + ((255 - a) * b)) div 255;
          sl[ndx + 1] := ((sl[ndx + 1] * a) + ((255 - a) * g)) div 255;
          sl[ndx + 2] := ((sl[ndx + 2] * a) + ((255 - a) * r)) div 255;
        end;
        // if a = 255 then do not place any color from the image at this
        // pixel, but leave the background intact instead.
      end;
    end;
  end;
end;

// -----------------------------------------------------------------------------
//
//			InitializeDemData
//
// -----------------------------------------------------------------------------
procedure TPngFile.InitializeDemData;
var
  cvaluep:  ^Cardinal;
  y:        Integer;
begin
  // Initialize Data and RowPtrs
  if Data <> nil then
    FreeMem(Data);
  Data := nil;
  if RowPtrs <> nil then
    FreeMem(RowPtrs);
  RowPtrs := nil;

  GetMem(Data, FHeight * FWidth * FBytesPerPixel);
  GetMem(RowPtrs, sizeof(Pointer) * FHeight);
  if (Data <> nil) and (RowPtrs <> nil) then
  begin
    cvaluep := Pointer(RowPtrs);
    for y := 0 to FHeight - 1 do
    begin
      cvaluep^ :=  Cardinal(Data) + Cardinal((FWidth * FBytesPerPixel * y)); //changed signed
      Inc(cvaluep);
    end;
  end;
end;

// -----------------------------------------------------------------------------
//
//			SaveToFile
//
// -----------------------------------------------------------------------------
procedure TPngFile.SaveToFile(const Filename: string);
var
  png:      PPng_Struct;
  pnginfo:  PPng_Info;
  tmp:      array[0..32] of char;
  pngfile:  Pointer;
begin
  pngfile := png_open_file(@Filename[1], 'wb');
  if pngfile = nil then
  begin
    if GFXRaiseErrors then  raise Exception.Create('Error opening PNG-file ' + Filename + '!');
    GFXFileErrorList.Add('Error opening PNG-file ' + Filename + '!');
    exit;
  end;

  try
    StrPCopy(tmp, PNG_LIBPNG_VER_STRING);
    try
      png := png_create_write_struct(tmp, nil, nil, nil);
      if png <> nil then
      begin
        try
          // create info struct and init io functions
          pnginfo := png_create_info_struct(png);
          png_init_io(png, pngfile);
          png_set_write_status_fn(png, nil);
          // set image attributes, compression, etc...
          png_set_IHDR(png, pnginfo, FWidth, FHeight, FBitDepth, FColorType,
                       PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT,
                       PNG_FILTER_TYPE_DEFAULT);

          if (Data <> nil) and (RowPtrs <> nil) then
          begin
            // Swap 16 bit images from PC Format
            if FBitDepth = 16 then
              png_set_swap(png);
            // Write the image
            png_write_image(png, PPByte(RowPtrs));
            png_write_end(png, pnginfo);
          end;  // if buf <> nil
        finally
          png_destroy_write_struct(@png, @pnginfo);
        end;  // try pnginfo create
      end;  // png <> nil
    except
      if GFXRaiseErrors then  raise Exception.Create('Error writing PNG-file!');
      GFXFileErrorList.Add('Error writing PNG-file!');
    end;  // try png create

  finally
    png_close_file(pngfile);
  end;
end;


initialization
  TPicture.RegisterFileFormat('PNG','PNG-Format', TPNGBitmap);

finalization
   TPicture.UnRegisterGraphicClass(TPNGBitmap);

end.
